# Copyright 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import subprocess
import unittest

import testing.test_changebranch
import gvn.cmdline
import gvn.util
import gvn.wc

from gvn.subcommands.change import ChangeEditor
from testing import testcommon

class TestChange(testing.test_changebranch.TestCase):

  def GetContext(self, options=None):
    if options is None:
      options = []
    options.append('-q')
    self.ctx = gvn.cmdline.Context(
                             gvn.cmdline.OptionParser().parse_args(options)[0],
                             [])
    self.config = self.ctx.config = testcommon.GetConfig()
    self.pool = self.ctx.pool
    self.wc = self.ctx.wc
    testcommon.GetAuthBaton(self.config, self.pool, self.wc.project.repository)

  def testChangeEditor(self):
    # base class for mock Editors, below
    class base(object):
      def __init__(*args, **kwargs):
        pass

    # Initialize context and get a changebranch.
    self.GetContext()
    paths = testing.test_changebranch.MakeChanges(self.ctx.wc, self.pool)
    change1 = self.BranchAndAssert('Foo', paths, 'Foo desc')

    # Get status to pass to ChangeEditor.
    path_status = {}
    def status_cb(target, status):
      if gvn.wc.IsModified(status):
        path_status[self.ctx.wc.RelativePath(target)] = status
    self.ctx.wc.Open(write_lock=False)
    self.ctx.wc.Status([], recursive=True,
                       get_all=False, no_ignore=False, show_children=False,
                       callback=status_cb)

    # Test simple ChangeEditor, with no changes to the form.
    class editor(base):
      def Edit(self, buf, encoding, tmp_prefix='gvn.'):
        return buf
    (desc,
     changed_paths) = ChangeEditor(self.ctx, change1, desc='', action=None,
                                   path_status=path_status,
                                   changed_paths=[x.relative_path
                                               for x in change1.changed_paths],
                                   editor=editor(), encoding='utf-8')
    self.assertEquals(desc, '\n')
    self.assertEquals(changed_paths, path_status)

    # Now test inserting a blank line in the to-be-branched section.
    class editor(base):
      def Edit(self, buf, encoding, tmp_prefix='gvn.'):
        result = buf.split('\n')
        result.insert(12, '')
        return '\n'.join(result)
    (desc,
     changed_paths) = ChangeEditor(self.ctx, change1, desc='', action=None,
                                   path_status=path_status,
                                   changed_paths=[x.relative_path
                                               for x in change1.changed_paths],
                                   editor=editor(), encoding='utf-8')
    self.assertEquals(desc, '\n')
    self.assertEquals(changed_paths, path_status)

    # Now test removing a path from the form.
    class editor(base):
      def Edit(self, buf, encoding, tmp_prefix='gvn.'):
        result = buf.split('\n')
        result[0] = 'new message\nwith two lines'
        # Remove 'new'
        del result[12]
        return '\n'.join(result)
    (desc,
     changed_paths) = ChangeEditor(self.ctx, change1, desc='', action=None,
                                   path_status=path_status,
                                   changed_paths=[x.relative_path
                                               for x in change1.changed_paths],
                                   editor=editor(), encoding='utf-8')
    self.assertEquals(desc, 'new message\nwith two lines\n')
    self.assertEquals(sorted(changed_paths.iterkeys()),
                      sorted(x for x in  path_status.iterkeys()
                             if x != 'new'))

  def testHandle_GvnChange_OneFile(self):
    # r302 broke 'gvn change just-one-file', as ctx.path became
    # 'just-one-file' but Handle_GvnChange was expecting it to be ''
    # (i.e. just-one-file's parent dir).  Fixing Handle_GvnChange
    # leaves ctx.Notify using RelativePath('foo', 'foo') => '.' as the
    # path to print.  So let's make sure this all works.

    # Edit iota.
    fp = open('foo', 'w')
    fp.write('hi\n')
    fp.close()
    testcommon.sleepf(testcommon.RunSvn, 'add', '-q', 'foo')

    # Setup Context.
    options = ['-c', 'change1', 'foo']
    ctx = gvn.cmdline.Context(*gvn.cmdline.OptionParser().parse_args(options))
    ctx.config = testcommon.GetConfig()
    testcommon.GetAuthBaton(ctx.config, ctx.pool, ctx.project.repository)

    # Mock ctx._stderr to grab the output.
    lines = []
    class mock_file(object):
      def write(self, line):
        lines.append(line)
      def flush(self): pass
    ctx._stderr = mock_file()

    # Mock gvn.util.Editor just returning the contents
    class mock_editor(gvn.util.Editor):
      def Edit(self, buf, encoding, tmp_prefix=None):
        return buf

    # Now run and assert.
    gvn.subcommands.change.Handle_GvnChange(ctx, mock_editor)
    self.assertEquals(lines,
                      ['Adding         foo\n',
                       'Transmitting file data ', '.',
                       '\n', 'Changed unittest/change1@%d.\n'
                       % (ctx.project.repository.GetHead(),)])

  def testHandle_GvnChange_Issue20(self):
    os.chdir('greek-tree/A/D')

    # Edit rho.
    fp = open('G/rho', 'a')
    fp.write('hi\n')
    testcommon.sleepf(fp.close)

    # Setup Context.
    options = ['--non-interactive', '-c', 'change1', 'G/rho']
    ctx = gvn.cmdline.Context(*gvn.cmdline.OptionParser().parse_args(options))
    ctx.config = testcommon.GetConfig()
    testcommon.GetAuthBaton(ctx.config, ctx.pool, ctx.project.repository)

    # Mock ctx._stderr to grab the output.
    lines = []
    class mock_file(object):
      def write(self, line):
        lines.append(line)
      def flush(self): pass
    ctx._stderr = mock_file()

    # Now run and assert.
    gvn.subcommands.change.Handle_GvnChange(ctx)
    self.assertEquals(lines,
                      ['Sending        %s\n' % (os.path.join('G', 'rho'),),
                       'Transmitting file data ', '.',
                       '\n', 'Changed unittest/change1@%d.\n'
                       % (ctx.project.repository.GetHead(),)])

    # Edit pi.
    fp = open('G/pi', 'a')
    fp.write('hi\n')
    testcommon.sleepf(fp.close)

    # Setup Context.
    options = ['--non-interactive', '-c', 'change1', 'G/pi']
    ctx = gvn.cmdline.Context(*gvn.cmdline.OptionParser().parse_args(options))
    ctx.config = testcommon.GetConfig()
    testcommon.GetAuthBaton(ctx.config, ctx.pool, ctx.project.repository)

    # Mock ctx._stderr to grab the output.
    lines = []
    class mock_file(object):
      def write(self, line):
        lines.append(line)
      def flush(self): pass
    ctx._stderr = mock_file()

    # Now run and assert.
    gvn.subcommands.change.Handle_GvnChange(ctx)
    self.assertEquals(lines,
                      ['Sending        %s\n' % (os.path.join('G', 'pi'),),
                       'Sending        %s\n' % (os.path.join('G', 'rho'),),
                       'Transmitting file data ', '.', '.',
                       '\n', 'Changed unittest/change1@%d.\n'
                       % (ctx.project.repository.GetHead(),)])
